package SliMP3::Playlist;

# $Id: Playlist.pm,v 1.166 2003/03/11 05:56:33 blackketter Exp $
# SliMP3 Server Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2.

use strict;

use File::Spec::Functions qw(:ALL);
use FileHandle;
use IO::Socket qw(:DEFAULT :crlf);
use Time::HiRes;
use SliMP3::Display;
use SliMP3::Misc;
use SliMP3::Scan;
use SliMP3::Strings qw(string);

my $CLIENTBUFFERLEN = (128 * 1024);

my %executeCommandCallbacks;

#############################################################################
# executecommand - does all the hard work.  Use it.
# takes:
#   a client reference
#   a reference to an array of parameters
#   a reference to a callback function
#   a list of callback function args
#
# returns an array containing the given parameters

sub executecommand {
	my($client, $parrayref, $callbackf, $callbackargs) = @_;
	my($item);
	my $callcallback = 1;
	my $p0 = $$parrayref[0];
	my $p1 = $$parrayref[1];
	my $p2 = $$parrayref[2];
	my $p3 = $$parrayref[3];
	my $p4 = $$parrayref[4];
	my $p5 = $$parrayref[5];

	$::d_playlist && msg(" Executing command " . $client->id .": $p0 (" .
			(defined $p1 ? $p1 : "") . ") (" .
			(defined $p2 ? $p2 : "") . ") (" .
			(defined $p3 ? $p3 : "") . ") (" .
			(defined $p4 ? $p4 : "") . ") (" .
			(defined $p5 ? $p5 : "") . ")\n");

	# The first parameter is the client identifier to execute the command.
	# If a parameter is "?" it is replaced by the current value in the array
	# returned by the function
	# Below are a list of the commands:
	#
	# P0	    P1		    P2							P3			P4		P5
	# play
	# pause 	(0|1|)
	# stop
	# mode		<play|pause|stop>
	# mode		?
	# sleep 	(0..n)																
	# sleep 	?																	
	# gototime	(0..n sec)|(-n..+n sec)
	# power 	(0|1|?)
	# time		?
	# genre		?
	# artist	?
	# album		?
	# title		?
	# duration  ?
	# playlist 	play 	    <song>
	# playlist 	append 	    <song>
	# playlist 	load 	    <playlist>
	# playlist 	add 	    <playlist>
	# playlist  loadalbum   <genre>						<artist>	<album> <song>
	# playlist  addalbum <genre> 					<artist>	<album> <song>
	# playlist 	clear
	# playlist 	move 		<fromoffset> 				<tooffset>
	# playlist 	delete 		<songoffset>
	# playlist 	jump 		<index>
	# playlist 	index		<index>						?
	# playlist	genre		<index>						?
	# playlist	artist		<index>						?
	# playlist	album		<index>						?
	# playlist	title		<index>						?
	# playlist	duration	<index>						?
	# playlist	tracks		?
	# mixer		volume		(0 .. 100)|(-100 .. +100)
	# mixer		volume		?
	# mixer		balance		(-100 .. 100)|(-200 .. +200)							(not implemented!)
	# mixer		bass		(0 .. 100)|(-100 .. +100)								
	# mixer		treble		(0 .. 100)|(-100 .. +100)								
	# display   <line1> 	<line2>                     (duration)
	# display   ? 			?
	# button   	<buttoncode>
	# player	count		?
	# player	name		<playerindex>				?
	# player	address		<playerindex>				?
	# pref		<prefname>	<prefvalue>
	# pref		<prefname>	?	
	# playerpref		<prefname>	<prefvalue>
	# playerpref		<prefname>	?	
	# rescan
	
	if (!defined($p0)) {
		# ignore empty commands
	# these commands don't require a valid client
	} elsif ($p0 eq "player") {
		if (!defined($p1)) {
		
		} elsif ($p1 eq "count") {
			$p2 = SliMP3::Client::clientCount();
		} elsif ($p1 eq "name") {
			my @clients = SliMP3::Client::clientIPs();
			if (defined $p2 && defined $clients[$p2]) {
				my $selectedClient  = SliMP3::Client::getClient($clients[$p2]);
				$p3 = SliMP3::Client::name($selectedClient);
			}
		} elsif ($p1 eq "address") {
			my @clients = SliMP3::Client::clientIPs();
			if (defined $p2 && defined $clients[$p2]) {
				$p3 = $clients[$p2];
			}
		}
	} elsif ($p0 eq "pref") {
		if (defined($p2) && $p2 ne '?' && !$::nosetup) {
			SliMP3::Prefs::set($p1, $p2);
		}
		$p2 = SliMP3::Prefs::get($p1);
	} elsif ($p0 eq "rescan") {
		SliMP3::MusicFolderScan::startScan();
	} elsif ($p0 eq "debug") {
		if ($p1 =~ /^d_/) {
			my $debugsymbol = "::" . $p1;
			no strict 'refs';
			if (!defined($p2)) {
				$$debugsymbol = ! $$debugsymbol;
				$p2 = $$debugsymbol;

			} elsif ($p2 eq "?")  {
				$p2 = $$debugsymbol;
			} else {
				$$debugsymbol = $p2;
			}	
			use strict 'refs';
			if (!$p2) { $p2 = 0; }
		}
	#the following commands require a valid client to be specified
	} elsif ($client) {
		if ($p0 eq "playerpref") {
			if (defined($p2) && $p2 ne '?' && !$::nosetup) {
				SliMP3::Prefs::clientSet($client, $p1, $p2);
			}
			$p2 = SliMP3::Prefs::clientGet($client, $p1);
		} elsif ($p0 eq "play") {
			playmode($client, "play");
			rate($client,1);
		} elsif ($p0 eq "pause") {
			if (defined($p1)) {
				if ($p1 && $client->playmode eq "play") {
					playmode($client, "pause");
				} elsif (!$p1 && $client->playmode eq "pause") {
					playmode($client, "resume");
				}
			} else {
				if ($client->playmode eq "pause") {
					playmode($client, "resume");
				} elsif ($client->playmode eq "play") {
					playmode($client, "pause");
				}
			}
		} elsif ($p0 eq "rate") {
			if ($client->mp3filehandleIsSocket) {
				rate($client, 1);
			} elsif (!defined($p1) || $p1 eq "?") {
				$p1 = rate($client);
			} else {
				if ($p1 == 0) {
					playmode($client, "pausenow");
				} elsif ($p1 == 1) {
					playmode($client, "play");
				}
				rate($client,$p1);
			}
		} elsif ($p0 eq "stop") {
			playmode($client, "stop");
		} elsif ($p0 eq "mode") {
			if (!defined($p1) || $p1 eq "?") {
				$p1 = $client->playmode
			} else {
				if ($client->playmode eq "pause" && $p1 eq "play") {
					playmode($client, "resume");
				} else {
					playmode($client, $p1);
				}
			}
		} elsif ($p0 eq "sleep") {
	
			if ($p1 eq "?") {
				$p1 = $client->sleepTime() - Time::HiRes::time();
				if ($p1 < 0) {
					$p1 = 0;
				}
			} else {
				SliMP3::Timers::killTimers($client, \&SliMP3::Playlist::gotosleep);
				
				if ($p1 != 0) {
					SliMP3::Timers::setTimer($client, Time::HiRes::time() + $p1, \&SliMP3::Playlist::gotosleep);
					$client->sleepTime(Time::HiRes::time() + $p1);
				} else {
					$client->sleepTime(0);
				}
			}	
		} elsif ($p0 eq "gototime" || $p0 eq "time") {
			if ($p1 eq "?") {
				$p1 = SliMP3::Playlist::songTime($client);
			} else {
				gototime($client, $p1, 1);
			}
		} elsif ($p0 eq "duration") {
			$p1 = SliMP3::Info::durationSeconds(SliMP3::Playlist::song($client)) || 0;
		} elsif ($p0 eq "artist") {
			$p1 = SliMP3::Info::artist(SliMP3::Playlist::song($client)) || 0;
		} elsif ($p0 eq "album") {
			$p1 = SliMP3::Info::album(SliMP3::Playlist::song($client)) || 0;
		} elsif ($p0 eq "title") {
			$p1 = SliMP3::Info::title(SliMP3::Playlist::song($client)) || 0;
		} elsif ($p0 eq "genre") {
			$p1 = SliMP3::Info::genre(SliMP3::Playlist::song($client)) || 0;
		} elsif ($p0 eq "path") {
			$p1 = SliMP3::Playlist::song($client) || 0;
		} elsif ($p0 eq "power") {
			if (!defined $p1) {
				SliMP3::Client::power($client, !SliMP3::Client::power($client));
			} elsif ($p1 eq "?") {
				$p1 = SliMP3::Client::power($client);
			} else {
				SliMP3::Client::power($client, $p1)
			}
		} elsif ($p0 eq "playlist") {
			if ($p1 =~ /^(?:add|append|load|play)$/) {
				$p2 = SliMP3::Misc::virtualToAbsolute($p2);
			}
			if ($p1 eq "play") {
				playmode($client, "stop");
				@{playList($client)} = (SliMP3::Misc::fixPath($p2));
				reshuffle($client);
				playmode($client, "play");
			} elsif ($p1 eq "append") {
				push(@{playList($client)}, SliMP3::Misc::fixPath($p2));
				reshuffle($client);
			} elsif ($p1 eq "load") {
				playmode($client, "stop");
				@{playList($client)} = ();
				SliMP3::Scan::addToList(playList($client), $p2, 1, undef, \&load_done, $client, $callbackf, $callbackargs);
				$callcallback = 0;
			} elsif ($p1 eq "add") {
				SliMP3::Scan::addToList(playList($client), $p2, 1, undef, \&add_done, $client, $callbackf, $callbackargs);
				$callcallback = 0;
			} elsif ($p1 eq "loadalbum") {
				playmode($client, "stop");
				@{playList($client)} = ();
				push(@{playList($client)}, SliMP3::Info::songs(singletonRef($p2), singletonRef($p3), singletonRef($p4), singletonRef($p5)));
				reshuffle($client);
				jumpto($client, 0);
			} elsif ($p1 eq "addalbum") {
				push(@{playList($client)}, SliMP3::Info::songs(singletonRef($p2), singletonRef($p3), singletonRef($p4), singletonRef($p5)));
				reshuffle($client);
			} elsif ($p1 eq "repeat") {
				# change between repeat values 0 (don't repeat), 1 (repeat the current song), 2 (repeat all)
				$client->htmlstatusvalid(0);
				if (!defined($p2)) {
					repeat($client, (repeat($client) + 1) % 3);
				} elsif ($p2 eq "?") {
					$p2 = repeat($client);
				} else {
					repeat($client, $p2);
				}
			} elsif ($p1 eq "shuffle") {
				if (!defined($p2)) {
					shuffle($client, !shuffle($client));
					reshuffle($client);
				} elsif ($p2 eq "?") {
					$p2 = shuffle($client);
				} else {
					shuffle($client, $p2);
					reshuffle($client);
				}
			} elsif ($p1 eq "clear") {
				@{playList($client)} = ();
				reshuffle($client);
				playmode($client, "stop");
			} elsif ($p1 eq "move") {
				#$p2 = from index
				#$p3 = to index
				my $listref;
				if (defined $p3 && $p3 =~ /^[\+-]/) {
					$p3 = $p2 + $p3;
				}
				if (defined $p2 && defined $p3 && $p2 < count($client) && $p3 < count($client) && $p2 >= 0 && $p3 >=0) {
					if (shuffle($client)) {
						$listref = shuffleList($client);
					} else {
						$listref = playList($client);
					}
					if (defined $listref) {
						$item = splice @{$listref},$p2, 1;
						splice @{$listref},$p3, 0, $item;
						my $currentSong = currentSongIndex($client);
						if ($p2 == $currentSong) {
							currentSongIndex($client,$p3);
						} elsif ($p3 == $currentSong) {
							currentSongIndex($client,($p3>$p2)? $currentSong - 1 : $currentSong + 1);
						}
						refreshPlaylist($client);
					}
				}
			} elsif ($p1 eq "delete") {
				if (defined($p2)) {
					remove($client, $p2);
				}
			} elsif (($p1 eq "jump") || ($p1 eq "index")) {
				if ($p2 eq "?") {
					$p2 = currentSongIndex($client);
				} else {
				#	if (playmode($client) eq 'play') {
				#		SliMP3::Control::fade_volume($client, -8, \&jumpto, [$client, $p2]);
				#	} else {
						jumpto($client, $p2);
				#	}
				}
			} elsif ($p1 eq "tracks") {
				$p2 = count($client);
			} elsif ($p1 eq "duration") {
				$p3 = SliMP3::Info::durationSeconds(SliMP3::Playlist::song($client,$p2)) || 0;
			} elsif ($p1 eq "artist") {
				$p3 = SliMP3::Info::artist(SliMP3::Playlist::song($client,$p2)) || 0;
			} elsif ($p1 eq "album") {
				$p3 = SliMP3::Info::album(SliMP3::Playlist::song($client,$p2)) || 0;
			} elsif ($p1 eq "title") {
				$p3 = SliMP3::Info::title(SliMP3::Playlist::song($client,$p2)) || 0;
			} elsif ($p1 eq "genre") {
				$p3 = SliMP3::Info::genre(SliMP3::Playlist::song($client,$p2)) || 0;
			} elsif ($p1 eq "path") {
				$p3 = SliMP3::Playlist::song($client,$p2) || 0;
			}
		} elsif ($p0 eq "mixer") {
			if ($p1 eq "volume") {
				my $newvol;
				my $oldvol = SliMP3::Prefs::clientGet($client, "volume");
				if ($p2 eq "?") {
					$p2 = $oldvol;
				} else {
					if($oldvol < 0) {
						# volume was previously muted
						$oldvol *= -1;      # un-mute volume
					} 
					
					if ($p2 =~ /^[\+\-]/) {
						$newvol = $oldvol + $p2;
					} else {
						$newvol = $p2;
					}
					
					if ($newvol> 100) { $newvol = $SliMP3::Control::maxVolume; }
					if ($newvol < 0) { $newvol = 0; }
					SliMP3::Prefs::clientSet($client, "volume", $newvol);
					SliMP3::Control::volume($client, $newvol);
				}
			} elsif ($p1 eq "muting") {
				my $vol = SliMP3::Prefs::clientGet($client, "volume");
				my $fade;
	
				if($vol < 0) {
					# need to un-mute volume
					$fade = 8;
				} else {
					# need to mute volume
					$fade = -8;
				}
				SliMP3::Control::fade_volume($client, $fade, \&SliMP3::Control::mute, [$client]);
			} elsif ($p1 eq "balance") {
				# unsupported yet
			} elsif ($p1 eq "treble") {
				my $newtreb;
				my $oldtreb = SliMP3::Prefs::clientGet($client, "treble");
				if ($p2 eq "?") {
					$p2 = $oldtreb;
				} else {
	
					if ($p2 =~ /^[\+\-]/) {
						$newtreb = $oldtreb + $p2;
					} else {
						$newtreb = $p2;
					}
					if ($newtreb > $SliMP3::Control::maxTreble) { $newtreb = $SliMP3::Control::maxTreble; }
					if ($newtreb < $SliMP3::Control::minTreble) { $newtreb = $SliMP3::Control::minTreble; }
					SliMP3::Prefs::clientSet($client, "treble", $newtreb);
					SliMP3::Control::treble($client, $newtreb);
				}
			} elsif ($p1 eq "bass") {
				my $newbass;
				my $oldbass = SliMP3::Prefs::clientGet($client, "bass");
				if ($p2 eq "?") {
					$p2 = $oldbass;
				} else {
	
					if ($p2 =~ /^[\+\-]/) {
						$newbass = $oldbass + $p2;
					} else {
						$newbass = $p2;
					}
					if ($newbass > $SliMP3::Control::maxBass) { $newbass = $SliMP3::Control::maxBass; }
					if ($newbass < $SliMP3::Control::minBass) { $newbass = $SliMP3::Control::minBass; }
					SliMP3::Prefs::clientSet($client, "bass", $newbass);
					SliMP3::Control::bass($client, $newbass);
				}
			}
		} elsif ($p0 eq "display") {
			if ($p1 eq "?" && $p2 eq "?") {
				my ($line1, $line2) = SliMP3::Display::curLines($client);
				$p1 = $line1;
				$p2 = $line2;
			} else {
				SliMP3::Buttons::ScreenSaver::wakeup($client);
				SliMP3::Animation::showBriefly($client, $p1, $p2, $p3);
			}
		} elsif ($p0 eq "button") {
			# all buttons now go through executecommand()
			SliMP3::IR::executeButton($client, $p1);
		}
	# mark html status page invalid so it will update
	$client->htmlstatusvalid(0);
	}	
	
	my @returnArray = ();
	
	if (defined($p0)) { push @returnArray, $p0 };
	if (defined($p1)) { push @returnArray, $p1 };
	if (defined($p2)) { push @returnArray, $p2 };
	if (defined($p3)) { push @returnArray, $p3 };
	if (defined($p4)) { push @returnArray, $p4 };
	if (defined($p5)) { push @returnArray, $p5 };
	
	$callcallback && $callbackf && (&$callbackf(@$callbackargs, @returnArray));

	executeCommandCallback($client, \@returnArray);
	
	$::d_playlist && msg(" Returning array: $p0 (" .
			(defined $p1 ? $p1 : "") . ") (" .
			(defined $p2 ? $p2 : "") . ") (" .
			(defined $p3 ? $p3 : "") . ") (" .
			(defined $p4 ? $p4 : "") . ") (" .
			(defined $p5 ? $p5 : "") . ")\n");
	
	return @returnArray;
}

sub setExecuteCommandCallback {
	my $callbackRef = shift;
	$executeCommandCallbacks{$callbackRef} = $callbackRef;
}

sub clearExecuteCommandCallback {
	my $callbackRef = shift;
	delete $executeCommandCallbacks{$callbackRef};
}

sub executeCommandCallback {
	my $client = shift;
	my $paramsRef = shift;

	no strict 'refs';
		
	foreach my $executecommandcallback (keys %executeCommandCallbacks) {
		$executecommandcallback = $executeCommandCallbacks{$executecommandcallback};
		&$executecommandcallback($client, $paramsRef);
	}
}

#
# accessors for playlist information
#
sub count {
	my $client = shift;
	return scalar(@{playList($client)});
}

sub song {
	my $client = shift;
	my $index = shift;
	
	if (count($client) == 0) {
		return;
	}

	if (!defined($index)) {
		$index = currentSongIndex($client);
	}
	return ${playList($client)}[${shuffleList($client)}[$index]];
}

sub shuffleList {
	my ($client) = shift;
	
	if (isSlave($client)) {
		$client = $client->master;
	}
	
	return $client->shufflelist;
}

sub playList {
	my ($client) = shift;

	if (isSlave($client)) {
		$client = $client->master;
	}
	
	return $client->playlist;
}

sub currentSongIndex {
	my $client = shift;
	my $newindex = shift;
	
	if (isSlave($client)) {
		$client = $client->master;
	}
	
	if (defined($newindex)) {
		$client->currentsong($newindex);
	}
	
	return $client->currentsong;
}

sub copyPlaylist {
	my $toclient = shift;
	my $fromclient = shift;

	@{$toclient->playlist} = @{$fromclient->playlist};
	@{$toclient->shufflelist} = @{$fromclient->shufflelist};
	$toclient->currentsong(	$fromclient->currentsong);	
	SliMP3::Prefs::clientSet($toclient, "shuffle", SliMP3::Prefs::clientGet($fromclient, "shuffle"));
	SliMP3::Prefs::clientSet($toclient, "repeat", SliMP3::Prefs::clientGet($fromclient, "repeat"));
}

sub remove {
	my $client = shift;
	my $tracknum = shift;
	
	my $playlistIndex = ${shuffleList($client)}[$tracknum];

	my $stopped = 0;
	my $oldmode = playmode($client);
	
	if ($tracknum == currentSongIndex($client)) {
		playmode($client, "stop");
		$stopped = 1;
	} elsif ($tracknum < currentSongIndex($client)) {
		currentSongIndex($client,currentSongIndex($client) - 1);
	}
	
	splice(@{playList($client)}, $playlistIndex, 1);

	my @reshuffled;
	my $counter = 0;
	foreach my $i (@{shuffleList($client)}) {
		if ($i < $playlistIndex) {
			push @reshuffled, $i;
		} elsif ($i > $playlistIndex) {
			push @reshuffled, ($i - 1);
		} else {
		}
	}
	
	if (isSlave($client)) {
		$client = $client->master;
	}
	
	@{$client->shufflelist} = @reshuffled;

	if ($stopped && ($oldmode eq "play")) {
		jumpto($client, $tracknum);
	}
	
	refreshPlaylist($client);

}

sub shuffle {
	my $client = shift;
	my $shuffle = shift;
	
	if (isSlave($client)) {
		$client = $client->master;
	}

	if (defined($shuffle)) {
		SliMP3::Prefs::clientSet($client, "shuffle", $shuffle);
	}
	
	return SliMP3::Prefs::clientGet($client, "shuffle") ? 1 : 0;
}

sub repeat {
	my $client = shift;
	my $repeat = shift;
	
	if (isSlave($client)) {
		$client = $client->master;
	}

	if (defined($repeat)) {
		SliMP3::Prefs::clientSet($client, "repeat", $repeat);
	}
	
	return SliMP3::Prefs::clientGet($client, "repeat");
}

sub songTime {
	my ($client) = shift;

	if (isSlave($client)) {
		$client = $client->master;
	}

	if ($client->mp3filehandleIsSocket) {
		my $startTime = $client->remoteStreamStartTime;
		if ($startTime) {
			return Time::HiRes::time() - $startTime;
		} else {
			return 0;
		}
	}

	my $size = $client->songtotalbytes;
	my $duration = $client->songduration;
	my $byterate;
	if (!$duration) {
		$byterate = 0;
	} else {
		$byterate = $size / $duration;
	}

	my $songtime;
	if (!$size) {
		$songtime = 0;
	} else {
		$songtime = (songRealPos($client) / $size * $duration);
	}

	$::d_playlist && msg( "byterate = $byterate, size = $size, songpos = ".songRealPos($client)." songtime = $songtime\n");

	return($songtime);

}	

# the "real" position in the track, accounting for client and server-side buffering
# TODO include server-side buffering
# TODO this won't work when scanning in reverse
#
sub songRealPos {
	my ($client) = shift;

	if (isSlave($client)) {
		$client = $client->master;
	}

	my $realpos = 0;
	if (defined($client->songpos) && defined(SliMP3::Stream::fullness($client))) {
		$realpos = $client->songpos - SliMP3::Stream::fullness($client);
	}

	if ($realpos<0) {
#		warn("came up with a negative position in the stream: ".
#		     "songpos = ".$client->songpos.", fullness = ".SliMP3::Stream::fullness($client)."\n");
		$realpos = 0;
	}
#	$::d_playlist && msg("songRealPos songPos: ". $client->songpos . " realpos: $realpos\n");
	return $realpos;	
}

#
# playlist synchronization routines
#
sub syncname {
	my $client = shift;
	my @buddies = syncedWith($client);
	if (isMaster($client)) {
		unshift @buddies , $client;
	} else {
		push @buddies , $client;
	}
	my @names = map {SliMP3::Client::name($_) || SliMP3::Client::id($_)} @buddies;
	$::d_playlist && msg("syncname for " . $client->id . " is " . (join ' & ',@names) . "\n");
	return join ' & ',@names;
}

# unsync a client from its buddies
sub unsync {
	my $client = shift;
	
	$::d_sync && msg( $client->id . ": unsyncing\n");
	
	# bail if we aren't synced already
	if (!isSynced($client)) {
		return;
	}
	
	# if we're the master...
	if (isMaster($client)) {
		my $slave;
		my $newmaster;
		
		# make a new master
		$newmaster = splice @{$client->slaves}, 0, 1;

		# you are your own master now
		$newmaster->master(undef);

		# if there are any slaves left
		if (scalar(@{$client->slaves}) > 0) {
		
			# make the slaves know about the new master
			foreach $slave (@{$client->slaves}) {
				$slave->master($newmaster);
			}		
			
			# copy over the slave list to the new master
			@{$newmaster->slaves} = @{$client->slaves};
			
		}
						
		# forget about our slaves
		@{$client->slaves} = ();

		# and copy the playlist to the new master
		copyPlaylist($newmaster, $client);	
		$newmaster->mp3filehandle($client->mp3filehandle);
		$client->mp3filehandle(undef);	
	} elsif (isSlave($client)) {
		# if we're a slave, remove us from the master's list
		my $i = 0;
		foreach my $c (@{($client->master())->slaves}) {
			if ($c->id eq $client->id) {
				splice @{$client->master->slaves}, $i, 1;
				last;
			}
			$i++;
		}	
	
		# and copy the playlist to the now freed slave
		my $master = $client->master;
		copyPlaylist($client, $master);
	
		$client->master(undef);
	}
	# when we unsync, we stop.
	executecommand($client, ["stop"]);
}

# sync a given client to another client
sub sync {
	my $client = shift;
	my $buddy = shift;
	
	$::d_sync && msg($client->id .": syncing\n");

	if (isSynced($client) && isSynced($buddy) && master($client) eq master($buddy)) {
		return;  # we're already synced up!
	}
	
	unsync($client);
	
	if (isSynced($buddy)) {
		$buddy = master($buddy);
	}
	
	$client->master($buddy);
	
	push (@{$client->master->slaves}, $client);
	
	if (playmode($buddy) eq "play") {
		SliMP3::Playlist::executecommand($client, ["playlist", "jump", "+0"]);
	}
	executeCommandCallback($client, ['playlist','sync']);

}

sub syncedWith {
	my $client = shift;
	my @buddies = ();
	my $otherclient;
	
	# get the master and its slaves
	if (isSlave($client)) {
		push @buddies, $client->master;
		foreach $otherclient (@{$client->master()->slaves}) {
			next if ($client == $otherclient);	# skip ourself
			push @buddies, $otherclient;
			$::d_sync && msg($client->id .": is synced with other slave " . $otherclient->id . "\n");
		}
	}
	
	# get our slaves
	foreach $otherclient (@{$client->slaves()}) {
		push @buddies, $otherclient;
		$::d_sync && msg($client->id . " : is synced with its slave " . $otherclient->id . "\n");
	}
	
	return @buddies;
}

sub isSyncedWith {
	my $client = shift;
	my $buddy = shift;
	
	foreach my $i (syncedWith($client)) {
		if ($buddy == $i) {
			$::d_sync && msg($client->id . " : is synced with " . $buddy->id . "\n");
			return 1;
		}
	}
	$::d_sync && msg($client->id . " : is synced NOT with " . $buddy->id . "\n");
	return 0;
}

sub isSynced {
	my $client = shift;
	return (isMaster($client) || isSlave($client));
}

sub canSyncWith {
	my $client = shift;
	my @buddies = ();
#	if (SliMP3::Client::isSliMP3($client)) {
		foreach my $otherclient (SliMP3::Client::clients()) {
			next if ($client eq $otherclient);				# skip ourself
#			next if (!SliMP3::Client::isSliMP3($otherclient));  	# we only sync SliMP3 devices
			next if (isSlave($otherclient)); 				# only include masters and un-sync'ed clients.
			push @buddies, $otherclient;
		}
#	}
	
	return @buddies;
}

# checkSync:
#   syncs up the start of playback of synced clients
#   resyncs clients between songs if some clients have multiple outstanding chunks
#
sub checkSync {
	my $client = shift;
	
	$::d_playlist_v && msg("checkSync: Player " . $client->id . " has " . scalar(@{$client->chunks}) . " chunks, and " . $client->usage() . "% full buffer \n");

	if (!SliMP3::Playlist::isSynced($client)) {
		return;
	}
	
	# if we're synced and waiting for the group's buffers to fill,
	# check if our buffer has passed the 95% level. If so, indicate
	# that we're ready to be unpaused.  If everyone else is now ready,
	# unpause all the clients at once.
	if ($client->readytosync == 0) {

		my $usage = $client->usage();
		$::d_playlist && msg("checking buffer usage: $usage on client $client\n");

		if 	($usage > 0.90) {
			$client->readytosync(1);
		
			$::d_playlist && msg($client->id." is ready to sync ".Time::HiRes::time()."\n");
			my $allReady=1;
			my $everyclient;
			foreach $everyclient ($client, SliMP3::Playlist::syncedWith($client)) {
				if (!($everyclient->readytosync)) {
					$allReady=0;
				}
			}
			
			if ($allReady) {
				$::d_playlist && msg("all clients ready to sync now. unpausing them.\n");
				foreach $everyclient ($client, SliMP3::Playlist::syncedWith($client)) {
					SliMP3::Control::resume($everyclient);
				}
			}
		}
	}
	
	my $everyclient;
	my @group = ($client, SliMP3::Playlist::syncedWith($client));
	
	# if we are ready to resync, then check to see if we've run out of packets.
	# when we have, stop and restart the clients on the current song, which 
	# has already been opened.
	
	if (master($client)->resync) {
		my $readyToContinue = 0;
		# we restart the song as soon as the first player has run out of chunks.
		foreach $everyclient (@group) {
			$::d_playlist && msg("Resync: Player " . $everyclient->id . " has " . scalar(@{$everyclient->chunks}) . " chunks \n");
			if (scalar(@{$everyclient->chunks}) == 0) { 
				$readyToContinue = 1; 
				last; 
			}
		}

		if ($readyToContinue) {
			master($client)->resync(0);
			SliMP3::Playlist::executecommand($client, ["playlist", "jump", "+0"]);
			$::d_playlist && msg("Resync restarting players on current song\n");
		}
	}
	
	# sanity check on queued chunks
	foreach my $everyclient (@group) {
		if (scalar(@{$everyclient->chunks}) > 200) { 
			$::d_playlist && msg("Player " . $everyclient->id . " isn't keeping up with the rest of the synced group.");
			@{$everyclient->chunks} = ();
			last; 
		}
	}
}

sub forgetClient {
	my $client = shift;
	# clear out the playlist
	executecommand($client, ["playlist", "clear"]);
	
	# trying to play will close out any open files.
	executecommand($client, ["play"]);
}

sub gotosleep {
	my $client = shift;
	if (SliMP3::Client::isSliMP3($client)) {
		SliMP3::Buttons::Common::setMode($client,'off');
	}
	$client->sleepTime(0);
	$client->currentSleepTime(0);
}

sub nextChunk {
	my $client = shift;
	my $maxChunkSize = shift;
	my $chunkRef;
	my $i;
		
	# if there's a chunk in the queue, then use it.
	if (scalar(@{$client->chunks})) {
		$chunkRef = shift @{$client->chunks};
	} else {
		#otherwise, read a new chunk
		my $readfrom = $client;
		
		if (isSlave($client)) {
			$readfrom = master($client);
		}

		# if we're at a non-regular rate (not 0 or 1), play for one second 
		# then skip a calculated amount to play back at the apparent rate.
		my $rate = rate($client);
		my $lastskip = $client->lastskip();
		my $now = Time::HiRes::time();
		if (($rate != 0 && $rate != 1) && (($now - $lastskip) > 1.0)) { 
			my $skip = $rate - 1.0;
			if ($skip > 0) { $skip = "+" . $skip; }
			$::d_playlist && msg("non regular play rate ($rate), skipping $skip seconds\n");
			$client->lastskip($now);
			gototime($readfrom, $skip, 1);
			# the gototime above will reload the data, so we should just return;
			return undef;
		}
			
		$chunkRef = readNextChunk($readfrom, $maxChunkSize);
			
		if (defined($chunkRef)) {	
			# let everybody I'm synced with use this chunk
			foreach my $buddy (syncedWith($client)) {
				push @{$buddy->chunks}, $chunkRef;
			}
		}
	}
	if (defined($chunkRef)) {
		my $len = length($$chunkRef);
		
		if ($len > $maxChunkSize) {
			push @{$client->chunks}, \substr($$chunkRef, $maxChunkSize - $len, $len - $maxChunkSize);
			$chunkRef = \substr($$chunkRef, 0, $maxChunkSize);
		}
		
		# remember the last chunk we sent
		if (defined($chunkRef)) {
			$client->lastchunk($chunkRef);
		}
	}
	
	return $chunkRef;
}

sub lastChunk {
	my $client = shift;
	return $client->lastchunk;
}


################################################################################
# Private functions below. Do not call from outside this module.
################################################################################

# openNext returns 1 if it succeeded opening a new song, 
#                  0 if it stopped at the end of the list or failed to open the current song 
# note: Only call this with a master or solo client
sub openNext {
	$::d_playlist && msg("opening next song...\n"); 
	
	my $client = shift;
	my $result = 1;
	
	# we're at the end of a song, let's figure out which song to open up.
	# if we can't open a song, skip over it until we hit the end or a good song...
	do {
	
		if (repeat($client) == 2  && $result) {
			# play the next song and start over if necessary
			skipsong($client);
		} elsif (repeat($client) == 1 && $result) {
			#play the same song again
		} else {
			#stop at the end of the list
			if (currentSongIndex($client) == (count($client) - 1)) {
				playmode($client, $result ? 'playout' : 'stop');
				return 0;
			} else {
				skipsong($client);
			}
		}
		
		$result = openSong($client);
		
	} while (!$result);
	
	# this is
	if ($result && isSynced($client)) {
		my $silence = SliMP3::HTTP::getStaticContent("html/silentpacket.mp3");
		my $count = int($CLIENTBUFFERLEN / length($silence)) + 1;
		my @fullbufferofsilence =  (\$silence) x $count;
		$::d_playlist && msg("stuffing " . scalar(@fullbufferofsilence) . " of silence into the buffers to sync.\n"); 
		# stuff silent packets to fill the buffers for each player
		foreach my $buddy ($client, syncedWith($client)) {
			push @{$buddy->chunks}, (@fullbufferofsilence);
		}
		$client->resync(1); 
	}
	
	return $result;
}

sub closeSong {
	my $client = shift;

	# close the previous handle to prevent leakage.
	if (defined $client->mp3filehandle()) {
		$client->mp3filehandle->close();
		$client->mp3filehandle(undef);
		$client->mp3filehandleIsSocket(0);
	}	
}

sub openSong {
	my $client = shift;

	# at the end of a song, reset the song time
	$client->songtotalbytes(0);
	$client->songduration(0);
	$client->songpos(0);

	# reset shoutcast variables
	$client->shoutMetaInterval(0);
	$client->shoutMetaPointer(0);

	# mark htmlplaylist invalid so the current song changes
	$client->htmlstatusvalid(0);
	my $fullpath = '';

	# We are starting a new song, lets kill any animation so we see the correct new song.
	foreach my $everyclient ($client, syncedWith($client)) { 
		SliMP3::Animation::killAnimation($everyclient);
	}
	
	closeSong($client);
	
	$fullpath = song($client);

	unless ($fullpath) {
		return undef;
	}


	$::d_playlist && msg("openSong on: $fullpath\n");
	####################
	# parse the filetype

	if (SliMP3::Info::isHTTPURL($fullpath)) {
		my $line1 = string('CONNECTING_FOR');
		my $line2 = SliMP3::Info::standardTitle($client, song($client));			
		SliMP3::Animation::showBriefly($client, $line1, $line2);

		# we don't get the content type until after the stream is opened
		my $sock = SliMP3::RemoteStream::openRemoteStream($fullpath, $client);
		if ($sock) {
			# if it's an mp3 stream, then let's stream it.
			if (SliMP3::Info::isSong($fullpath)) {
				$client->mp3filehandle($sock);
				$client->mp3filehandleIsSocket(1);
			# if it's one of our playlists, parse it...
			} elsif (SliMP3::Info::isList($fullpath)) {
				$::d_playlist && msg("openSong on a remote list!\n");
				my @items;
				# handle the case that we've actually got a playlist in the list,
				# rather than a stream.

				# parse out the list
				@items = SliMP3::Parse::parseList($fullpath, $sock);
				
				# hack to preserve the title of a song redirected through a playlist
				if (  scalar(@items) == 1 && 
				      (SliMP3::Info::title($items[0]) eq $items[0]) &&
				      defined(SliMP3::Info::title($fullpath))) {
				    SliMP3::Info::addTitleToCache($items[0], SliMP3::Info::title($fullpath));
				} 
				
				
				# close the socket
				$sock->close();
				$sock = undef;
				$client->mp3filehandle(undef);

				# insert the list onto the playlist
				splice @{playList($client)}, currentSongIndex($client), 1, @items;

				# update the shuffle list
				reshuffle($client);

				# try to open the first item in the list, if there is one.
				return openSong($client);
			} else {
				$::d_playlist && msg("don't know how to handle content for $fullpath\n");
				$sock->close();
				$sock = undef;
				$client->mp3filehandle(undef);
			}
		} 
		
		if (!$sock) {
			$::d_playlist && msg("Remote stream failed to open, showing message.\n");
			$client->mp3filehandle(undef);
			
			my $line1 = string('PROBLEM_CONNECTING');
			my $line2 = SliMP3::Info::standardTitle($client, song($client));			
			SliMP3::Animation::showBriefly($client, $line1, $line2, 5);
			return undef;
		}
	} elsif (SliMP3::Info::isMP3($fullpath) || 
	         (SliMP3::Info::isFileURL($fullpath) && SliMP3::Info::isSong($fullpath))) {
		$::d_playlist && msg("openSong: this is an MP3 file: $fullpath\n");
		
		my $origpath = $fullpath;
		if (SliMP3::Info::isFileURL($fullpath) && SliMP3::Info::isSong($fullpath)) {
			$fullpath = SliMP3::Misc::pathFromFileURL($fullpath);
		}
		my $size = SliMP3::Info::size($origpath);
		my $duration = SliMP3::Info::durationSeconds($origpath);
		my $offset = SliMP3::Info::offset($origpath);

		$::d_playlist && msg("openSong: getting duration  $duration, size $size, and offset $offset for $origpath\n");
		
		if (!$size || !$duration)  {
			$::d_playlist && msg("openSong: not bothering opening file with zero size or duration\n");
			return undef;
		}
		$client->mp3filehandle( FileHandle->new() );

		$::d_playlist && msg("openSong: opening file $fullpath\n");
		if ($client->mp3filehandle->open($fullpath)) {
			
			$client->songtotalbytes($size);
			$client->songduration($duration);
			$client->songoffset($offset);
			
			$::d_playlist && msg(" seeking in $offset into $fullpath\n");
			if ($offset) {
				if (!seek ($client->mp3filehandle, $offset, 0) ) {
					msg("couldn't seek to $offset for $fullpath");
				};
			}
						
		} else { 
			$client->mp3filehandle(undef);
		}
	} elsif (SliMP3::Info::isOgg($fullpath)) {

		# Note we have to put the path in double quotes so that
		# spaces in file names are handled properly.
		# Only tested on Linux.
		my $ogg_cmd = "ogg123 -q -p 5 -d raw -f - \"$fullpath\"";
		my $lame_cmd = "lame -x -r - - &";
		
		$client->mp3filehandle( FileHandle->new() );

		$client->mp3filehandle->open("$ogg_cmd | $lame_cmd |");
        $client->songtotalbytes(SliMP3::Info::size($fullpath));
        $client->songduration(SliMP3::Info::durationSeconds($fullpath));

	} elsif( SliMP3::Info::isWav( $fullpath)) {
	        my $samplerate = SliMP3::Prefs::get("wavmp3samplerate");
		my $lame_cmd = qq(lame --silent -h -b $samplerate "$fullpath" -) . (SliMP3::OSDetect::OS() eq 'win' ? "" : " &");

		$client->mp3filehandle( FileHandle->new() );
		$client->mp3filehandle->open( "$lame_cmd |");
	        $client->songtotalbytes(SliMP3::Info::size($fullpath));
        	$client->songduration(SliMP3::Info::durationSeconds($fullpath));

	} else {
		$::d_playlist && msg("Next item is of unrecognized type! Stopping! $fullpath\n");
		return undef;
	}

	######################
	# make sure the filehandle was actually set
	if ($client->mp3filehandle()) {
		binmode($client->mp3filehandle());
		SliMP3::History::record(song($client));
	} else {
		$::d_playlist && msg("Can't open [$fullpath] : $!");

		my $line1 = string('PROBLEM_OPENING');
		my $line2 = SliMP3::Info::standardTitle($client, song($client));		
		SliMP3::Animation::showBriefly($client, $line1, $line2, 5);
		return undef;
	}

	refreshPlaylist($client);
	
	executeCommandCallback($client,  ['open', $fullpath]);

	return 1;
}

sub refreshPlaylist {
	my $client = shift;
	# make sure we're displaying the new current song in the playlist view.
	foreach my $everybuddy ($client, syncedWith($client)) {
		if (SliMP3::Client::isSliMP3($everybuddy)) {
			SliMP3::Buttons::Playlist::jump($everybuddy);
		}
		$client->htmlstatusvalid(0); #invalidate cached htmlplaylist
	}
	
}

# skipsong is just for playing the next song when the current one ends
sub skipsong {
	my ($client) = @_;
	# mark htmlplaylist invalid so the current song changes
	$client->htmlstatusvalid(0);

	currentSongIndex($client, currentSongIndex($client) + 1);

	if (currentSongIndex($client) >= count($client)) {
		if (shuffle($client) && SliMP3::Prefs::get('reshuffleOnRepeat')) {
			my $playmode = playmode($client);
			playmode($client,'stop');
			reshuffle($client);
			playmode($client,$playmode);
		}
		currentSongIndex($client, 0);
	}
}

sub readNextChunk {
	my $client = shift;
	my $givenChunkSize = shift;
	
	if (!defined($givenChunkSize)) {
		$givenChunkSize = SliMP3::Prefs::get('udpChunkSize') * 10;
	} 
	
	my $chunksize = $givenChunkSize;
	
	my $chunk  = 0;

	if ($client->mp3filehandle()) {
	
		my $isRemoteStream = $client->mp3filehandleIsSocket;
		
		if ($isRemoteStream) {
			# If the MP3 file handle is a remote stream and it's not readable,
			# just return instead of blocking here. The client will repeat the
			# request.
			#
			my $selRead = IO::Select->new();
			$selRead->add($client->mp3filehandle);
			my ($selCanRead,$selCanWrite)=IO::Select->select($selRead,undef,undef,0);
			if (!$selCanRead) {
				$::d_playlist && msg("remote stream not readable\n");
				return undef;
			} else {
				$::d_playlist && msg("remote stream readable.\n");
			}

			# adjust chunksize to lie on metadata boundary (for shoutcast/icecast)
			if ($client->shoutMetaInterval() &&
				($client->shoutMetaPointer() + $chunksize) > $client->shoutMetaInterval()) {
	
				$chunksize = $client->shoutMetaInterval() - $client->shoutMetaPointer();
			}
		} else {
			# don't send extraneous ID3 data at the end of the file
			my $size = $client->songtotalbytes();
			my $pos = $client->songpos();
			
			if ($pos + $chunksize > $size) {
				$chunksize = $size - $pos;
				$::d_playlist && msg( "Reduced chunksize to $chunksize at end of file");
			}
		}

		$client->mp3filehandle()->read($chunk, $chunksize);
		
		if ($isRemoteStream) {
			$client->shoutMetaPointer($client->shoutMetaPointer() + length($chunk));
	
			# handle instream metadata for shoutcast/icecast
			if ($client->shoutMetaPointer() == $client->shoutMetaInterval()) {
	
				SliMP3::RemoteStream::readMetaData($client);
				$client->shoutMetaPointer(0);
			}
		}
	} else {
		$::d_playlist && msg($client->id . ": No filehandle to read from, returning no chunk.\n");
		$::d_playlist && bt();
		return undef;
	}

	# if nothing was read from the filehandle, then we're done with it,
	# so open the next filehandle.
	if (length($chunk) == 0 ||
		($client->songtotalbytes() != 0 && ($client->songpos()) > $client->songtotalbytes())) {
		$::d_playlist && msg("opening next song: chunk length" . length($chunk) . ", song pos: " .
				$client->songpos() . "(tell says: . " . tell($client->mp3filehandle()). "), totalbytes: " . $client->songtotalbytes() . "\n");
		if (!openNext($client)) {
			$::d_playlist && msg("Can't opennext, returning no chunk.");
			return undef;
		}
		
		$client->mp3filehandle()->read($chunk, $givenChunkSize);
		
		if ($client->mp3filehandleIsSocket) {
			$client->shoutMetaPointer(length($chunk));
		}
	}
	
	$client->songpos($client->songpos + length($chunk));
	
	$::d_playlist_verbose && msg("read a chunk of " . length($chunk) . " length\n");
	
	return \$chunk;
}

# rate can be negative for rew, zero for pause, 1 for playback and greater than one for ffwd
sub rate {
	my ($client, $newrate) = @_;
	if (!defined($newrate)) {
		return $client->rate;
	}
	my $oldrate = $client->rate();
	$client->rate($newrate);
	
	$::d_playlist && msg("switching rate from $oldrate to $newrate\n");
	# restart playback if we've changed and we're not pausing or unpauseing
	if ($oldrate != $newrate) {  		
	 	if ($oldrate != 0 && $newrate != 0) {
	 		$::d_playlist && msg("gototime, jumping to here in order to restart the stream\n");
			gototime($client, "+0", 1);
		}
	}
}

# playmode - start playing, pause or stop
sub playmode {
	my($client, $newmode) = @_;

	if (!defined($newmode)) {
		return $client->playmode;
	}
	
	$::d_playlist && $newmode && msg($client->id . ": Switching to mode $newmode\n");

	my $prevmode = $client->playmode;

	if ($newmode eq $prevmode) {
		$::d_playlist && msg(" Already in playmode $newmode : ignoring mode change\n");
	} else {

		# if we're playing, then open the new song the master.		
		if ($newmode eq "play") {
			my $opened;
			
			# if the player is off, we automatically power on when we start to play
			if (!SliMP3::Client::power($client)) {
				SliMP3::Client::power($client, 1);
			}
			
			if (isSlave($client)) {
				$opened = openSong(master($client));
			} else {
				$opened = openSong($client);
			}
			# if we couldn't open the song, then stop...
			if (!$opened) {
				$::d_playlist && msg("Couldn't open song.  Stopping.\n");
				$newmode = "stop";
			}
		}
		
		# when we change modes, make sure we do it to all the synced clients.
		foreach my $everyclient ($client, syncedWith($client)) {
			$::d_playlist && msg(" New play mode: " . $newmode . "\n");
	
			# when you resume, you go back to play mode
			if (($newmode eq "resume") ||($newmode eq "resumenow")) {
				$everyclient->playmode("play");
				rate($everyclient, 1);
				
			} elsif ($newmode eq "pausenow") {
				$everyclient->playmode("pause");
				rate($everyclient, 0);
				
			} elsif ($newmode eq "playout") {
				$everyclient->playmode("stop");
				
			} else {
				$everyclient->playmode($newmode);
			}
	
			if ($newmode eq "stop") {
				$everyclient->currentplayingsong("");
			#	$everyclient->songpos(0);
				$::d_playlist && msg("Stopping and clearing out old chunks for client " . $everyclient->id . "\n");
				@{$everyclient->chunks} = ();

				SliMP3::Control::stop($everyclient);
				closeSong($everyclient);
			} elsif ($newmode eq "play") {
				$everyclient->readytosync(0);
				rate($everyclient, 1);
				SliMP3::Control::play($everyclient, SliMP3::Playlist::isSynced($everyclient));				
			} elsif ($newmode eq "pause") {
				# since we can't count on the accuracy of the fade timers, we unfade them all, but the master calls back to unpause everybody
				if ($everyclient eq $client) {
					SliMP3::Control::fade_volume($everyclient, -8, \&pauseSynced, [$client]);
				} else {
					SliMP3::Control::fade_volume($everyclient, -8);
				}				
				
			} elsif ($newmode eq "pausenow") {
				SliMP3::Control::pause($everyclient);
				rate($everyclient,0);
			} elsif ($newmode eq "resumenow") {
				SliMP3::Control::resume($everyclient);
				rate($everyclient, 1);
				
			} elsif ($newmode eq "resume") {
				# set volume to 0 to make sure fade works properly
				SliMP3::Control::volume($everyclient,0);
				SliMP3::Control::resume($everyclient);
				SliMP3::Control::fade_volume($everyclient, 8);
				
			} elsif ($newmode eq "playout") {
				SliMP3::Control::playout($everyclient);
				
			} else {
				$::d_playlist && msg(" Unknown play mode: " . $everyclient->playmode . "\n");
				return $everyclient->playmode();
			}
			
			refreshPlaylist($everyclient);
		}
	}
	$::d_playlist && msg($client->id . ": Current playmode: " . $client->playmode() . "\n");

	return $client->playmode();
}

sub pauseSynced {
	my $client = shift;
	foreach my $everyclient ($client, syncedWith($client)) {
		SliMP3::Control::pause($everyclient);
		rate($everyclient, 0);
	}
}

# jumpto - set the current song to a given offset
sub jumpto {
	my($client, $offset) = @_;
	my($songcount) = count($client);
	if ($songcount == 0) {
		return;
	}

	playmode($client,"stop");
	if ($offset =~ /[\+\-]\d+/) {
		currentSongIndex($client, currentSongIndex($client) + $offset);
		$::d_playlist && msgf("jumping by %s\n", $offset);
	} else {
		currentSongIndex($client, $offset);
		$::d_playlist && msgf("jumping to %s\n", $offset);
	}

	if ($songcount && currentSongIndex($client) >= $songcount) {
		currentSongIndex($client, currentSongIndex($client) % $songcount);
	}
	if ($songcount && currentSongIndex($client) < 0) {
		currentSongIndex($client, $songcount - ((0 - currentSongIndex($client)) % $songcount));
	}

	playmode($client,"play");
}

#reshuffle - every time the playlist is modified, the shufflelist should be updated
#		We also invalidate the htmlplaylist at this point
sub reshuffle {
	my($client) = @_;
	my($realsong);
	my($i);
	my($temp);
	my($songcount) = count($client);
	my $listRef = shuffleList($client);

	$client->htmlstatusvalid(0); #invalidate cached htmlplaylist

	if ($songcount) {
		$realsong = ${$listRef}[currentSongIndex($client)];

		if (!defined($realsong)) {
			$realsong = 0;
		} elsif ($realsong > $songcount) {
			$realsong = $songcount;
		}

		@{$listRef} = (0 .. ($songcount - 1));

		if (shuffle($client)) {
			for ($i = $songcount; --$i; ) {
				# swap each item with a random item;
				$a = int(rand($i + 1));
				@$listRef[$i,$a] = @$listRef[$a,$i];
			}
		}

		if (playmode($client) eq "stop") {
			currentSongIndex($client,0);
		} else {
			for ($i = 0; $i < $songcount; $i++) {
				if (${$listRef}[$i] == $realsong) {
					if (shuffle($client)) {
						$temp = ${$listRef}[$i];
						${$listRef}[$i] = ${$listRef}[0];
						${$listRef}[0] = $temp;
						$i = 0;
					}
					currentSongIndex($client,$i);
					last;
				}
			}
		}
		

	} else {
		@{$listRef} = ();
		currentSongIndex($client, 0);
	}
	refreshPlaylist($client);
}

#
# jump to a particular time in the current song
#  should be dead-on for CBR, approximate for VBR
#  third argument determines whether this is an instant jump or wait until the
#   buffer gets around to it
#
sub gototime {
	my($client, $newtime, $doitnow) = @_;
	my $newoffset;
	my $oldoffset;
	my $oldtime;

	if (isSynced($client)) {
		$client = master($client);
	}

	return if (!song($client));
	return if !defined($client->mp3filehandle);

	my $size = $client->songtotalbytes;
	my $duration = $client->songduration;

#	$oldoffset = $client->songpos;
	$oldoffset = songRealPos($client);
	
	if ($oldoffset > $size) {
		 warn ("song position > size\n") 
	}

	$oldtime = songTime($client);

	if ($newtime =~ /^[\+\-]/) {
		$newoffset = int($oldoffset + ($newtime / $duration * $size));
		$::d_playlist && msgf("going to time %s, offset %s from old offset %s\n", $newtime, $newoffset, $oldoffset);
		# skip to the previous or next track as necessary
		if ($newoffset > $size) {
			my $rate = rate($client);
			jumpto($client, "+1");
			rate($client, $rate);
			$newtime = ($newoffset - $size) * $duration / $size;
			$::d_playlist && msg("skipping forward to the next track to time $newtime\n");
			gototime($client, $newtime, $doitnow);
			return;
		} elsif ($newoffset < 0) {
			my $rate = rate($client);
			while ($newtime < 0) {
				jumpto($client, "-1");
				rate($client, $rate);
				$newtime = $client->songduration - ((-$newoffset) * $duration / $size);
				$::d_playlist && msg("skipping backwards to the previous track to time $newtime\n");
			}
			gototime($client, $newtime, $doitnow);
			return;
		}
	} else {
		$::d_playlist && msgf("going to time %s\n", $newtime);
		$newoffset = int($newtime / $duration * ($size));
		if ($newoffset > $size) {
			$newoffset = $size;
		} elsif ($newoffset < 0) {
			$newoffset = 0;
		}
	}
	
	$::d_playlist && msgf("oldoffset $oldoffset, old time = %d, newoffset: %d new time = %d\n", $oldtime, $newoffset, $newoffset / $size * $duration);

	if ($doitnow) {
		foreach my $everybuddy ($client, slaves($client)) {
			$::d_playlist && msg("gototime: stopping playback\n");
			SliMP3::Control::stop($everybuddy);
			@{$everybuddy->chunks} = ();
		}
	}
	my $dataoffset =  $client->songoffset;
	$client->songpos($newoffset);
	$client->mp3filehandle->seek($newoffset+$dataoffset, 0);

	if ($doitnow) {
		foreach my $everybuddy ($client, slaves($client)) {
			$::d_playlist && msg("gototime: restarting playback\n");
			$everybuddy->readytosync(0);
			SliMP3::Control::play($everybuddy, SliMP3::Playlist::isSynced($client));
		}
	}	
}

sub isMaster {
	my $client = shift;
	if (scalar(@{$client->slaves}) > 0) {
		return 1;
	} else {
		return 0;
	}
}

sub master {
	my $client = shift;
	if (isMaster($client)) {
		return $client;
	} 
	return $client->master;
}

sub slaves {
	my $client = shift;
	
	return @{$client->slaves};
}

sub isSlave {
	my $client = shift;
	if (defined($client->master)) {
		return 1;
	} else {
		return 0;
	}
}

sub load_done {
	my ($client, $callbackf, $callbackargs)=@_;
	reshuffle($client);
	jumpto($client, "0");
	$callbackf && (&$callbackf(@$callbackargs));
	executeCommandCallback($client, ['playlist','load_done']);
}

sub add_done {
	my ($client, $callbackf, $callbackargs)=@_;
	reshuffle($client);
	$callbackf && (&$callbackf(@$callbackargs));
	executeCommandCallback($client, ['playlist','add_done']);
}

sub singletonRef {
    my $arg = shift;

	if (!defined($arg)) {
		return [];
    } elsif ($arg eq '*') {
        return [];
    } elsif ($arg) {
        return [$arg];
    } else {
        return [];
    }
}


1;
__END__

# Local Variables:
# tab-width:4
# indent-tabs-mode:t
# End:
